from typing import List
import numpy as np

from .bblock.vocabulary import Vocabulary


class BinaryTF:
    def __init__(self, vocabulary: Vocabulary):
        self.vocabulary = vocabulary

    def __call__(self, tokens: List[str]):
        v = np.zeros(len(self.vocabulary))

        for token in tokens:
            t = self.vocabulary[token]
            if not t.is_oov:
                v[t.id] = 1

        return v


class RawTF:
    def __init__(self, vocabulary: Vocabulary):
        self.vocabulary = vocabulary

    def __call__(self, tokens: List[str]):
        v = np.zeros(len(self.vocabulary))

        for token in tokens:
            t = self.vocabulary[token]
            if not t.is_oov:
                v[t.id] += 1

        return v

# def get_tf_vectorizer_by_name(name: str, vocabulary: Vocabulary):
#     if name is None:
#         name = configuration['tf']
#
#     if name == "binary":
#         pass
#     elif name == "raw":
#         pass
#     elif name == "freq":
#         pass
#     elif name == "log_norm":
#         pass
#     elif name == "double_norm":
#         pass
#     else:
#         warnings.warn(f"Unknown tf method {name}. Failing back to binary tf", UserWarning)
#
#         return BinaryTF(vocabulary)
#
#
# def get_idf_vectorizer_by_name(name: str):
#     pass



__title__ = "sadedegel"  # pragma: no cover
__version__ = "0.18.3"  # pragma: no cover
__release__ = True  # pragma: no cover
__download_url__ = "https://github.com/globalmaksimum/sadedegel/releases"  # pragma: no cover
__herokuapp_url__ = "https://sadedegel.herokuapp.com"  # pragma: no cover



from math import ceil
from os.path import dirname
from pathlib import Path

import numpy as np
from joblib import dump, load as jl_load
from rich.console import Console
from sklearn.linear_model import PassiveAggressiveClassifier
from sklearn.metrics import f1_score
from sklearn.model_selection import KFold
from sklearn.utils import shuffle

from ..dataset.tweet_sentiment import load_tweet_sentiment_train, CORPUS_SIZE, CLASS_VALUES
from ..extension.sklearn import TfidfVectorizer, OnlinePipeline

console = Console()


def empty_model():
    return OnlinePipeline(
        [('tfidf', TfidfVectorizer(tf_method='freq', idf_method='smooth', show_progress=True)),
         ('pa', PassiveAggressiveClassifier(C=6.438004890835467e-10, average=True))
         ]
    )


def cv(k=3):
    try:
        import pandas as pd
    except ImportError:
        console.log(("pandas package is not a general sadedegel dependency."
                     " But we do have a dependency on building our prebuilt models"))

    raw = load_tweet_sentiment_train()
    df = pd.DataFrame.from_records(raw)
    df = shuffle(df)

    BATCH_SIZE = 1000

    kf = KFold(n_splits=k)
    console.log(f"Corpus Size: {CORPUS_SIZE}")

    scores = []

    for train_indx, test_index in kf.split(df):
        train = df.iloc[train_indx]
        test = df.iloc[test_index]

        n_split = ceil(len(train) / BATCH_SIZE)
        console.log(f"{n_split} batches of {BATCH_SIZE} instances...")

        batches = np.array_split(train, n_split)

        pipeline = empty_model()

        for batch in batches:
            pipeline.partial_fit(batch.tweet, batch.sentiment_class,
                                 classes=[i for i in range(len(CLASS_VALUES))])

        y_pred = pipeline.predict(test.tweet)

        scores.append(f1_score(test.sentiment_class, y_pred))

        console.log(scores)


def build():
    try:
        import pandas as pd
    except ImportError:
        console.log(("pandas package is not a general sadedegel dependency."
                     " But we do have a dependency on building our prebuilt models"))

    raw = load_tweet_sentiment_train()
    df = pd.DataFrame.from_records(raw)
    df = shuffle(df)

    BATCH_SIZE = 1000

    console.log(f"Corpus Size: {CORPUS_SIZE}")

    n_split = ceil(len(df) / BATCH_SIZE)
    console.log(f"{n_split} batches of {BATCH_SIZE} instances...")

    batches = np.array_split(df, n_split)

    pipeline = empty_model()

    for batch in batches:
        pipeline.partial_fit(batch.tweet, batch.sentiment_class,
                             classes=[i for i in range(len(CLASS_VALUES))])

    console.log("Model build [green]DONE[/green]")

    model_dir = Path(dirname(__file__)) / 'model'

    model_dir.mkdir(parents=True, exist_ok=True)

    dump(pipeline, (model_dir / 'tweet_sentiment.joblib').absolute(), compress=('gzip', 9))


def load(model_name="tweet_sentiment"):
    return jl_load(Path(dirname(__file__)) / 'model' / f"{model_name}.joblib")


if __name__ == '__main__':
    build()



from itertools import tee

import numpy as np
from rich.progress import track
from scipy.sparse import csr_matrix
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.pipeline import Pipeline

from ..config import config_context


def check_type(X):
    if not all(isinstance(x, str) for x in X):
        raise ValueError(f"X should be an iterable string (documents)")


class OnlinePipeline(Pipeline):
    def partial_fit(self, X, y=None, **kwargs):
        for i, step in enumerate(self.steps):
            name, est = step
            est.partial_fit(X, y, **kwargs)
            if i < len(self.steps) - 1:
                X = est.transform(X)
        return self


class SadedegelVectorizer(BaseEstimator, TransformerMixin):
    """Sadedegel feature extraction TransformerMixin s don't require fit calls."""

    def fit(self, X, y=None):
        return self

    def partial_fit(self, X, y=None, **kwargs):
        return self


class TfidfVectorizer(SadedegelVectorizer):
    def __init__(self, *, tf_method='raw', idf_method='probabilistic', drop_stopwords=True, lowercase=True,
                 drop_suffix=True, drop_punct=True, show_progress=True):
        super().__init__()

        self.tf_method = tf_method
        self.idf_method = idf_method
        self.lowercase = lowercase
        self.drop_suffix = drop_suffix
        self.drop_stopwords = drop_stopwords
        self.drop_punct = drop_punct
        self.show_progress = show_progress

        self.Doc = None

    def transform(self, X, y=None):
        if isinstance(X, list):
            check_type(X)
            n_total = len(X)
        else:
            X1, X2, X = tee(X, 3)

            check_type(X1)
            n_total = sum((1 for _ in X2))

        if n_total == 0:
            raise ValueError(f"Ensure that X contains at least one valid document. Found {n_total}")

        if self.Doc is None:
            with config_context(tokenizer="bert") as Doc:
                self.Doc = Doc

        indptr = [0]
        indices = []
        data = []
        for doc in track(X, total=n_total, description="Transforming document(s)", update_period=1,
                         disable=not self.show_progress):
            d = self.Doc(doc)
            n_vocabulary = len(d.builder.tokenizer.vocabulary)
            tfidf = d.get_tfidf(self.tf_method, self.idf_method, drop_stopwords=self.drop_stopwords,
                                lowercase=self.lowercase,
                                drop_suffix=self.drop_suffix,
                                drop_punct=self.drop_punct)

            for idx in tfidf.nonzero()[0]:
                indices.append(idx)
                data.append(tfidf[idx])

            indptr.append(len(indices))

        return csr_matrix((data, indices, indptr), dtype=np.float32, shape=(n_total, n_vocabulary))


class BM25Vectorizer(SadedegelVectorizer):
    def __init__(self, *, tf_method='raw', idf_method='probabilistic', k1=1.25, b=0.75, delta=0, drop_stopwords=True,
                 lowercase=True, drop_suffix=True, drop_punct=True, show_progress=True):

        super().__init__()

        self.tf_method = tf_method
        self.idf_method = idf_method
        self.lowercase = lowercase
        self.drop_suffix = drop_suffix
        self.drop_stopwords = drop_stopwords
        self.drop_punct = drop_punct
        self.show_progress = show_progress
        self.k1 = k1
        self.b = b
        self.delta = delta

        self.Doc = None

    def transform(self, X, y=None):
        if isinstance(X, list):
            check_type(X)
            n_total = len(X)
        else:
            X1, X2, X = tee(X, 3)

            check_type(X1)
            n_total = sum((1 for _ in X2))

        if n_total == 0:
            raise ValueError(f"Ensure that X contains at least one valid document. Found {n_total}")

        if self.Doc is None:
            with config_context(tokenizer="bert") as Doc:
                self.Doc = Doc

        indptr = [0]
        indices = []
        data = []
        for doc in track(X, total=n_total, description="Transforming document(s)", update_period=1,
                         disable=not self.show_progress):
            d = self.Doc(doc)
            n_vocabulary = len(d.builder.tokenizer.vocabulary)
            bm25 = d.get_bm25(self.tf_method, self.idf_method, drop_stopwords=self.drop_stopwords,
                              lowercase=self.lowercase,
                              drop_suffix=self.drop_suffix,
                              drop_punct=self.drop_punct,
                              k1=self.k1, b=self.b, delta=self.delta)

            for idx in bm25.nonzero()[0]:
                indices.append(idx)
                data.append(bm25[idx])

            indptr.append(len(indices))

        return csr_matrix((data, indices, indptr), dtype=np.float32, shape=(n_total, n_vocabulary))



from math import ceil
from os.path import dirname
from pathlib import Path

import numpy as np
from joblib import dump, load as jl_load
from rich.console import Console
from sklearn.linear_model import PassiveAggressiveClassifier
from sklearn.metrics import f1_score
from sklearn.utils import shuffle

from sadedegel.dataset.profanity import load_offenseval_train, load_offenseval_test, \
    load_offenseval_test_label, CORPUS_SIZE, CLASS_VALUES
from sadedegel.extension.sklearn import TfidfVectorizer, OnlinePipeline

console = Console()


def empty_model():
    return OnlinePipeline(
        [('tfidf', TfidfVectorizer(tf_method='freq', idf_method='smooth')),
         ('pa', PassiveAggressiveClassifier(C=8.770192177653856, average=True))])


def build():
    try:
        import pandas as pd
    except ImportError:
        console.log(("pandas package is not a general sadedegel dependency."
                     " But we do have a dependency on building our prebuilt models"))

    raw = load_offenseval_train()
    df = pd.DataFrame.from_records(raw)
    df = shuffle(df)

    BATCH_SIZE = 1000

    console.log(f"Corpus Size: {CORPUS_SIZE}")

    n_split = ceil(len(df) / BATCH_SIZE)
    console.log(f"{n_split} batches of {BATCH_SIZE} instances...")

    batches = np.array_split(df, n_split)

    pipeline = empty_model()

    for batch in batches:
        pipeline.partial_fit(batch.tweet, batch.profanity_class, classes=[i for i in range(len(CLASS_VALUES))])

    console.log("Model build [green]DONE[/green]")

    model_dir = Path(dirname(__file__)) / 'model'

    model_dir.mkdir(parents=True, exist_ok=True)

    dump(pipeline, (model_dir / 'tweet_profanity_classification.joblib').absolute(), compress=('gzip', 9))


def load():
    return jl_load(Path(dirname(__file__)) / 'model' / 'tweet_profanity_classification.joblib')


def evaluate():
    try:
        import pandas as pd
    except ImportError:
        console.log(("pandas package is not a general sadedegel dependency."
                     " But we do have a dependency on building our prebuilt models"))

    model = load()

    test = pd.DataFrame.from_records(load_offenseval_test())
    testLabel = pd.DataFrame.from_records(load_offenseval_test_label())

    test = test.merge(testLabel, on='id')

    y_pred = model.predict(test.tweet)

    console.log(f"Model test accuracy (f1-macro): {f1_score(test.profanity_class, y_pred, average='macro')}")


if __name__ == "__main__":
    build()



import csv
from pathlib import Path
from rich.console import Console

CLASS_VALUES = ['NOT', 'OFF']
CORPUS_SIZE = 31277

console = Console()

__general_download_message__ = """Ensure that you have properly downloaded profanity corpus using

            python -m sadedegel.dataset.profanity download --access-key xxx --secret-key xxxx
            
        Unfortunately due to data licensing issues we could not share data publicly. 
        Get in touch with sadedegel team to obtain a download key.
        """


def check_directory_structure(path: str) -> bool:
    base_dir = Path(path).expanduser()

    offenseval_dir = base_dir / 'profanity' / 'offenseval2020-turkish'

    if not base_dir.exists():
        console.log(f"Dataset base directory ([bold red]{base_dir}[/bold red]) does not exist")

    elif not offenseval_dir.exists():
        console.log(
            f"OffensEval directory ([bold red]{offenseval_dir}[/bold red]) does not exist")

    else:
        return True

    console.log(__general_download_message__)

    return False


def load_offenseval_train(data_home="~/.sadedegel_data"):
    if not check_directory_structure(data_home):
        raise Exception("Ts Corpus validation error")

    train_tsv = Path(data_home).expanduser() / "profanity" / "offenseval2020-turkish" / "offenseval-tr-training-v1"
    train_tsv = train_tsv / "offenseval-tr-training-v1.tsv"

    with open(train_tsv) as csvfile:
        rd = csv.DictReader(csvfile, delimiter='\t')

        for rec in rd:
            yield dict(id=int(rec['id']), tweet=rec['tweet'], profanity_class=CLASS_VALUES.index(rec['subtask_a']))


def load_offenseval_test(data_home="~/.sadedegel_data"):
    if not check_directory_structure(data_home):
        raise Exception("Ts Corpus validation error")

    test_tsv = Path(data_home).expanduser() / "profanity" / "offenseval2020-turkish" / "offenseval-tr-testset-v1"
    test_tsv = test_tsv / "offenseval-tr-testset-v1.tsv"

    with open(test_tsv) as csvfile:
        rd = csv.DictReader(csvfile, delimiter='\t')

        for rec in rd:
            yield dict(id=int(rec['id']), tweet=rec['tweet'])


def load_offenseval_test_label(data_home="~/.sadedegel_data"):
    if not check_directory_structure(data_home):
        raise Exception("Ts Corpus validation error")

    test_tsv = Path(data_home).expanduser() / "profanity" / "offenseval2020-turkish" / "offenseval-tr-testset-v1"
    test_tsv = test_tsv / "offenseval-tr-labela-v1.tsv"

    with open(test_tsv) as csvfile:
        rd = csv.DictReader(csvfile, fieldnames=['id', 'profanity_class'])

        for rec in rd:
            yield dict(id=int(rec['id']), profanity_class=CLASS_VALUES.index(rec['profanity_class']))



from typing import List
import numpy as np
import warnings
from collections import defaultdict
from os.path import dirname
from pathlib import Path

__tr_upper__ = "ABCÇDEFGĞHIİJKLMNOÖPRSŞTUÜVYZ"
__tr_lower__ = "abcçdefgğhıijklmnoöprsştuüvyz"

__tr_lower_abbrv__ = ['hz.', 'dr.', 'prof.', 'doç.', 'org.', 'sn.', 'st.', 'mah.', 'mh.', 'sok.', 'sk.', 'alb.', 'gen.',
                      'av.', 'ist.', 'ank.', 'izm.', 'm.ö.', 'k.k.t.c.']


def tr_lower(s: str) -> str:
    return s.replace("I", "ı").replace("İ", "i").lower()


def tr_upper(s: str) -> str:
    return s.replace("i", "İ").upper()


def space_pad(token):
    return " " + token + " "


def pad(l, padded_length):
    return l + [0 for _ in range(padded_length - len(l))]


def flatten(l2: List[List]):
    flat = []
    for l in l2:
        for e in l:
            flat.append(e)

    return flat


def is_eos(span, sentences: List[str]) -> int:
    start = 0
    eos = []
    for s in sentences:
        idx = span.doc.raw.find(s, start) + len(s) - 1
        eos.append(idx)

        start = idx

    b, e = span.value

    for idx in eos:
        if b <= idx <= e:
            return 1

    return 0


def select_layer(bert_out: tuple, layers: List[int], return_cls: bool) -> np.ndarray:
    """Selects and averages layers from BERT output.

    Parameters:
        bert_out: tuple
            Tuple containing output of 12 intermediate layers after feeding a document.

        layers: List[int]
            List that contains which layer to choose. max = 11, min = 0.

        return_cls: bool
            Whether to use CLS token embedding as sentence embedding instead of averaging token embeddings.

    Returns:
        numpy.ndarray (n_sentences, embedding_size) Embedding size if default to 768.

    """
    n_layers = len(layers)
    n_sentences = bert_out[0].shape[0]
    n_tokens = bert_out[0].shape[1]

    if not (min(layers) > -1 and max(layers) < 12):
        raise Exception(f"Value for layer should be in 0-11 range")

    if return_cls:
        cls_matrix = np.zeros((n_layers, n_sentences, 768))
        l_ix = 0
        for l, layer in enumerate(bert_out):
            if l not in layers:
                continue
            else:
                l_ix = l_ix + 1
            for s, sentence in enumerate(layer):
                cls_tensor = sentence[0].numpy()
                cls_matrix[l_ix - 1, s, :] = cls_tensor
        layer_mean_cls = np.mean(cls_matrix, axis=0)

        return layer_mean_cls

    else:
        token_matrix = np.zeros((n_layers, n_sentences, n_tokens - 2, 768))
        for l, layer in enumerate(bert_out):
            l_ix = 0
            if l not in layers:
                continue
            else:
                l_ix = l_ix + 1
            for s, sentence in enumerate(layer):
                for t, token in enumerate(sentence[1:-1]):  # Exclude [CLS] and [SEP] embeddings
                    token_tensor = sentence[t].numpy()
                    token_matrix[l_ix - 1, s, t, :] = token_tensor

        tokenwise_mean = np.mean(token_matrix, axis=2)
        layer_mean_token = np.mean(tokenwise_mean, axis=0)

        return layer_mean_token


def normalize_tokenizer_name(tokenizer_name, raise_on_error=False):
    normalized = tokenizer_name.lower().replace(' ', '').replace('-', '').replace('tokenizer', '')

    if normalized not in ['bert', 'simple']:
        msg = f"Invalid tokenizer {tokenizer_name} ({normalized}). Valid values are bert, simple"
        if raise_on_error:
            raise ValueError(msg)
        else:
            warnings.warn(msg, UserWarning, stacklevel=3)

    return normalized


def to_config_dict(kw: dict):
    d = defaultdict(lambda: dict())
    for k, v in kw.items():
        if '__' not in k:  # default section
            d['default'][k] = v
        else:
            section, key = k.split('__')

            d[section][key] = v

    return d


def load_stopwords(base_path=None):
    """ Return Turkish stopwords as list from file. """
    if base_path is None:
        base_path = dirname(__file__)

    text_path = Path(base_path) / "data" / "stop-words.txt"

    with open(text_path, "r") as fp:
        stopwords = fp.readlines()

    stopwords = [s.rstrip() for s in stopwords]

    return stopwords



import os.path
from pathlib import Path
import sys
from itertools import tee

from smart_open import open

import click

import boto3

from loguru import logger

from ._core import load_offenseval_test_label, load_offenseval_train, load_offenseval_test, CORPUS_SIZE, CLASS_VALUES

from zipfile import ZipFile

from rich.console import Console

console = Console()

logger.disable("sadedegel")


@click.group(help="Tokenization Dataset Commandline")
def cli():
    pass


@cli.command()
@click.option("--access-key", help="Access Key ID to download dataset.", prompt=True,
              default=lambda: os.environ.get('sadedegel_access_key', ''))
@click.option("--secret-key", help="Secret Key ID to download dataset.", prompt=True,
              default=lambda: os.environ.get('sadedegel_secret_key', ''))
@click.option("--data_home", '-d', help="Data home directory", default="~/.sadedegel_data")
def download(access_key, secret_key, data_home):
    """Download tokenization corpus from cloud with your key."""

    data_home = Path(os.path.expanduser(data_home)) / 'profanity'
    data_home.mkdir(parents=True, exist_ok=True)
    console.print(f"Data directory for profanity data {data_home}")

    transport_params = {
        'session': boto3.Session(aws_access_key_id=access_key,
                                 aws_secret_access_key=secret_key),
        'resource_kwargs': {
            'endpoint_url': 'https://storage.googleapis.com',
        }
    }

    url = f"s3://sadedegel/dataset/offenseval2020-turkish.zip"

    with open(url, 'rb', transport_params=transport_params) as fp:
        with ZipFile(fp) as zp:
            zp.extractall(data_home)


@cli.command()
def validate():
    """Sanity check on corpus
    """
    with console.status("[bold yellow]Validating train"):
        train = load_offenseval_train()

        train_clone, train = tee(train, 2)

        n_train = sum(1 for _ in train_clone)

        if n_train == CORPUS_SIZE:
            console.log("Cardinality check [yellow]DONE[/yellow]")
        else:
            console.log("Cardinality check [red]FAILED[/red]")
            console.log(f"|Profanity (train)| : {n_train}")
            sys.exit(1)

    with console.status("[bold yellow]Validate test"):
        test = set((d['id'] for d in load_offenseval_test()))
        test_label = set((d['id'] for d in load_offenseval_test_label()))

        a_b, ab, b_a = test - test_label, test & test_label, test_label - test

        if len(a_b) == 0 and len(b_a) == 0:
            console.log("Corpus check [green]DONE[/green]")
        else:
            console.log(f"Test file [red]DIVERGE[/red] from label file. {len(a_b)}, {len(b_a)}")


if __name__ == "__main__":
    cli()



from sqlalchemy import create_engine
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy import Column, Integer, String
from sqlalchemy import UniqueConstraint


engine = create_engine('sqlite:///score.db', echo=True)

Base = declarative_base()


class Configuration(Base):
    __table__ = "configuration"

    id = Column(Integer, primary_key=True)
    git_hash = Column(String, primary_key=True)
    git_hash_short = Column(String, primary_key=True)
    instance_str = Column(String, primary_key=True)

    UniqueConstraint('')




from typing import List
import numpy as np  # type: ignore
from ._base import ExtractiveSummarizer
from ..bblock import Sentences


class BM25Summarizer(ExtractiveSummarizer):
    """
    Assign a higher importance score based on BM25 score of the sentences within the document.

    k1 : Coefficient to be used for BM25 computation.
    b : Coefficient to be used for BM25 computation.
    delta: BM25+ term

    normalize : bool, optional (default=True)
        If ``False``, return a raw score vector.
        Otherwise, return L2 normalized score vector.
    """
    tags = ExtractiveSummarizer.tags + ['self-supervised', 'ml', 'ir', 'bm']

    def __init__(self, tf_method, idf_method, k1=1.25, b=0.75, delta=0,
                 drop_stopwords=False, drop_suffix=False,
                 drop_punct=False, lowercase=False, normalize=True):
        super().__init__(normalize)

        self.k1 = k1
        self.b = b
        self.delta = delta

        self.drop_stopwords = drop_stopwords
        self.drop_suffix = drop_suffix
        self.drop_punct = drop_punct
        self.lowercase = lowercase

        self.tf_method = tf_method
        self.idf_method = idf_method

    def _predict(self, sents: List[Sentences]):
        return np.array([sent.get_bm25(tf_method=self.tf_method,
                                       idf_method=self.idf_method,
                                       drop_stopwords=self.drop_stopwords,
                                       drop_suffix=self.drop_suffix,
                                       drop_punct=self.drop_punct,
                                       k1=self.k1, b=self.b, delta=self.delta).sum() for sent in sents])



import click
import numpy as np
from itertools import islice, tee

from collections import defaultdict

from rich.table import Table
from rich.console import Console
from rich.progress import track

from ...dataset.tscorpus import load_tokenization_tokenized, load_tokenization_raw, fail_with_corpus_validation, \
    CORPUS_SIZE
from ...dataset.extended import load_extended_sents_corpus
from ...config import tokenizer_context
from ..vocabulary import Vocabulary

console = Console()


def tok_eval(tokenizers=['simple', 'bert'], limit=None):
    """Evaluate tokenizers on a downsized version of tokenization dataset.

    :param tokenizers: List of tokenizer names to evaluate. Defaults to ["simple", "bert"].
    :type tokenizers: List[str]
    :param limit: Top-K documents for evaluation
    :type limit: int

    :return: IoU score between list of true tokens and list of tokenized tokens.
    """

    scores = []
    for tokenizer in tokenizers:

        console.log(f"[bold blue]Evaluating {tokenizer} tokenizer...")
        with tokenizer_context(tokenizer) as Doc:

            if limit:
                raw = islice(load_tokenization_raw(), limit)
                tok = islice(load_tokenization_tokenized(), limit)
            else:
                raw = load_tokenization_raw()
                tok = load_tokenization_tokenized()

            macro = []
            micro = defaultdict(int)
            for i, (raw_doc, tokenized_doc) in track(enumerate(zip(raw, tok)), description="Evaluating...",
                                                     total=limit if limit else CORPUS_SIZE):
                if raw_doc['id'] != tokenized_doc['id']:
                    fail_with_corpus_validation()

                true_tokens = set(tokenized_doc['tokens'])

                tokens = Doc(raw_doc['text']).tokens

                if tokenizer == 'bert':
                    fixed_hashtags = []
                    for tok in tokens:
                        if "##" not in tok:
                            fixed_hashtags.append(tok)
                        else:
                            merged = fixed_hashtags[-1] + tok.split("##")[1]
                            fixed_hashtags.pop(-1)
                            fixed_hashtags.append(merged)
                    tokens = fixed_hashtags

                pred_tokens = set(tokens)

                union = len(true_tokens.union(pred_tokens))
                intersection = len(true_tokens.intersection(pred_tokens))

                macro.append(intersection / union)

                micro['intersection'] += intersection
                micro['union'] += union

            scores.append((tokenizer, f"{np.mean(macro):.4f}", f"{micro['intersection'] / micro['union']:.4f}"))

    return scores


@click.group()
def cli():
    pass


@click.option("--tokenizers", "-t", help="List of tokenizers to evaluate", default=["bert", "simple"])
@click.option("--limit", "-s", help="Evaluation set size", type=int)
@cli.command()
def tokenizer_evaluate(tokenizers, limit):
    """Tokenizer evaluation"""

    result = tok_eval(tokenizers, limit)

    table = Table(show_header=True, header_style="bold magenta")
    table.add_column("Tokenizer")
    table.add_column("IoU (macro)")
    table.add_column("IoU (micro)")

    for row in result:
        table.add_row(*row)

    console.print(table)


@cli.command()
@click.option('--max-doc', help="Maximum documentation in extended corpus", type=int, default=-1)
@click.option('--min-df', help="Mininum document frequenct of a word to be included in", default=3)
@click.option('--word-tokenizer', type=click.Choice(['bert'], case_sensitive=False),
              help="Word tokenizer to be used in building vocabulary.", default='bert')
def build_vocabulary(max_doc, min_df, word_tokenizer):
    """Build vocabulary"""

    if max_doc > 0:
        corpus = islice(load_extended_sents_corpus(), max_doc)
    else:
        corpus = load_extended_sents_corpus()

    corpus, replica = tee(corpus, 2)
    total = sum(1 for _ in replica)

    vocab = Vocabulary.factory(word_tokenizer)

    click.secho(click.style(f"...Frequency calculation over extended dataset", fg="blue"))

    with tokenizer_context(word_tokenizer) as Doc:
        for i, d in track(enumerate(corpus), total=max_doc if max_doc > 0 else total,
                          description="Building vocabulary..."):
            doc = Doc.from_sentences(d['sentences'])

            for sent in doc:
                for word in sent.tokens:
                    vocab.add_word_to_doc(word, i)

    vocab.build(min_df)
    vocab.save()

    click.secho(click.style(f"Total documents {vocab.document_count}", fg="blue"))
    click.secho(click.style(f"Vocabulary size {len(vocab)} (words occurred more than {min_df} documents)", fg="blue"))


if __name__ == '__main__':
    cli()



from os.path import dirname
from pathlib import Path
import warnings
from dataclasses import dataclass, asdict
from json import dump, load
from collections import defaultdict

from .util import tr_lower, normalize_tokenizer_name
from .token import Token


@dataclass
class Entry:
    id: int
    word: str
    df: int
    df_cs: int
    vocabulary: object


class Vocabulary:
    vocabularies = {}

    @staticmethod
    def factory(tokenizer):
        normalized_name = normalize_tokenizer_name(tokenizer)

        if normalized_name not in Vocabulary.vocabularies:
            Vocabulary.vocabularies[normalized_name] = Vocabulary(normalized_name)

        return Vocabulary.vocabularies[normalize_tokenizer_name(tokenizer)]

    def __init__(self, tokenizer):
        self.tokenizer = tokenizer
        self.entries = {}
        self.initialized = False
        self.document_count = -1

        self.doc_counter = defaultdict(set)
        self.doc_counter_case_sensitive = defaultdict(set)
        self.doc_set = set()

        self.token_cache = {}

    @property
    def size(self):
        return len(self.entries)

    def __len__(self):
        return self.size

    def add_word_to_doc(self, word, doc_identifier):
        self.doc_counter[tr_lower(word)].add(doc_identifier)
        self.doc_counter_case_sensitive[word].add(doc_identifier)
        self.doc_set.add(doc_identifier)

    def build(self, min_df=1):
        i = 0
        for word in self.doc_counter_case_sensitive:
            if len(self.doc_counter[tr_lower(word)]) >= min_df:
                self.entries[word] = Entry(i, word, len(self.doc_counter[tr_lower(word)]),
                                           len(self.doc_counter_case_sensitive[word]), self)
                i += 1

        self.document_count = len(self.doc_set)
        self.initialized = True

    def save(self):
        if not self.initialized:
            raise Exception("Call build to initialize vocabulary")

        with open(Path(dirname(__file__)) / 'data' / 'vocabulary.json', "w") as fp:
            dump(dict(size=len(self), document_count=self.document_count, tokenizer=self.tokenizer,
                      words=[asdict(e) for e in self.entries.values()]), fp,
                 ensure_ascii=False)

    @staticmethod
    def load(tokenizer):
        normalized_name = normalize_tokenizer_name(tokenizer)

        vocab = Vocabulary.factory(normalized_name)

        if not vocab.initialized:
            if normalized_name != 'bert':
                warnings.warn("Currently only valid tokenizer is BERT Tokenizer for vocabulary generation.",
                              UserWarning, stacklevel=2)
                return vocab

            with open(Path(dirname(__file__)) / 'data' / 'vocabulary.json') as fp:
                json = load(fp)

            for d in json['words']:
                vocab.entries[d['word']] = Entry(d['id'], d['word'], d['df'], d['df_cs'], vocab)

            vocab.document_count = json['document_count']
            vocab.initialized = True

        return vocab

    def __getitem__(self, word):
        if not self.initialized:
            raise Exception(("Vocabulary instance is not initialize. "
                             "Use load for built in vocabularies, use build for manual vocabulary build."))
        else:
            token = self.token_cache.get(word, None)

            if token:
                return token
            else:
                entry = self.entries.get(word, None)

                if not entry:
                    entry = Entry(-1, word, 0, 0, self)  # OOV has a document frequency of 0 by convention

                self.token_cache[word] = Token(entry)

                return self.token_cache[word]
